import datetime
import json
from datetime import timezone
from enum import Enum
from typing import Any, ClassVar, Generic, Literal, TypeVar

import structlog
from llm_taxi import Message, Role
from llm_taxi.llms import LLM
from lopt import parse_model
from pydantic import BaseModel, ConfigDict, Field

from function_calling.function_calling.prompt import (
    DEFAULT_ARGUMENT_EXTRACTION_EXAMPLE_TEMPLATE,
    DEFAULT_ARGUMENT_EXTRACTION_PROMPT_TEMPLATE,
    DEFAULT_FUNCTION_SELECTION_EXAMPLE_TEMPLATE,
    DEFAULT_FUNCTION_SELECTION_FUNCTION_TEMPLATE,
    DEFAULT_FUNCTION_SELECTION_PROMPT_TEMPLATE,
)
from function_calling.function_calling.utils import json_schema_to_code

logger = structlog.get_logger()


class Function(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")

    name: str
    description: str
    parameters: dict[str, Any] = Field(default_factory=dict)


class NotSelected(BaseModel):
    pass


NOT_SELECTED = NotSelected()


class SelectedFunction(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")

    name: str
    arguments: str


class FunctionCallingOutput(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")

    type: Literal["function"] = "function"
    function: SelectedFunction | NotSelected


class FewShotExample(BaseModel):
    example: Any
    function: SelectedFunction


class Mode(str, Enum):
    Json = "json"
    Code = "code"


class FunctionSelectionResponse(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")

    function: str | None


class ArgumentExtractionResponse(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")

    arguments: dict[str, Any]


LLMType = TypeVar("LLMType")


class FunctionCalling(Generic[LLMType]):
    def __init__(  # noqa: PLR0913
        self,
        llm: LLM[LLMType],
        *,
        function_selection_prompt_template: str = DEFAULT_FUNCTION_SELECTION_PROMPT_TEMPLATE,
        function_selection_function_template: str = DEFAULT_FUNCTION_SELECTION_FUNCTION_TEMPLATE,
        function_selection_example_template: str = DEFAULT_FUNCTION_SELECTION_EXAMPLE_TEMPLATE,
        argument_extraction_prompt_template: str = DEFAULT_ARGUMENT_EXTRACTION_PROMPT_TEMPLATE,
        argument_extraction_example_template: str = DEFAULT_ARGUMENT_EXTRACTION_EXAMPLE_TEMPLATE,
    ) -> None:
        self.llm: LLM[LLMType] = llm

        self.function_selection_prompt_template: str = (
            function_selection_prompt_template
        )
        self.function_selection_function_template: str = (
            function_selection_function_template
        )
        self.function_selection_example_template: str = (
            function_selection_example_template
        )
        self.argument_extraction_prompt_template: str = (
            argument_extraction_prompt_template
        )
        self.argument_extraction_example_template: str = (
            argument_extraction_example_template
        )

    async def select_function(
        self,
        *,
        messages: list[Message],
        functions: list[Function],
        few_shot_examples: list[FewShotExample] | None = None,
        llm_options: dict[str, Any] | None = None,
    ) -> str | NotSelected | None:
        if not llm_options:
            llm_options = {
                "temperature": 0,
                "seed": 42,
            }

        functions_str = "\n\n".join(
            self.function_selection_function_template.format(
                name=x.name,
                description=x.description,
            )
            for x in functions
        )

        if self.function_selection_example_template and few_shot_examples:
            examples = [
                self.function_selection_example_template.format(
                    input=x.example,
                    output=next(f.name for f in functions if f.name == x.function.name),
                )
                for x in few_shot_examples
            ]
            examples_str = "\n\n".join(examples)
        else:
            examples_str = ""

        content = self.function_selection_prompt_template.format(
            input=messages[-1].content,
            functions=functions_str,
            examples=examples_str,
        )
        logger.debug("Formatted function selection prompt template", content=content)

        messages = [Message(role=Role.User, content=content)]
        response = await self.llm.response(messages, llm_options=llm_options)
        logger.debug("LLM response for function selection", response=response)

        if function_selection := parse_model(
            response, lang="json", model=FunctionSelectionResponse
        ):
            if function_selection.function is not None:
                logger.info("Selected function", function=function_selection.function)
                return function_selection.function

            logger.info("Function not selected")

            return NOT_SELECTED

        logger.warning("Failed to parse function selection response")

        return None

    async def extract_arguments(
        self,
        *,
        messages: list[Message],
        function: Function,
        few_shot_examples: list[FewShotExample] | None = None,
        mode: Mode = Mode.Json,
        llm_options: dict[str, Any] | None = None,
    ) -> dict[str, Any] | None:
        if not llm_options:
            llm_options = {
                "temperature": 0,
                "seed": 42,
            }

        datetime_str = datetime.datetime.now(tz=timezone.utc).isoformat()

        if self.argument_extraction_example_template and few_shot_examples:
            few_shot_examples = [
                x for x in few_shot_examples if x.function.name == function.name
            ]

            examples_str = "\n\n".join(
                [
                    self.argument_extraction_example_template.format(
                        input=x.example,
                        output=x.function.arguments,
                    )
                    for x in few_shot_examples
                ],
            )
        else:
            examples_str = ""

        if mode == Mode.Json:
            function_str = function.model_dump_json()
        else:
            function_str = json_schema_to_code(
                {
                    "type": "object",
                    "title": function.name,
                    "description": function.description,
                    "properties": function.parameters["properties"],
                },
            )

        content = self.argument_extraction_prompt_template.format(
            function=function_str,
            input=messages[-1].content,
            datetime=datetime_str,
            function_name=function.name,
            examples=examples_str,
        )
        logger.info("Formatted argument extraction prompt template", content=content)

        messages = [Message(role=Role.User, content=content)]
        response = await self.llm.response(messages, llm_options=llm_options)
        logger.debug("LLM response for argument extraction", response=response)

        if arguments := parse_model(
            response,
            lang="json",
            model=ArgumentExtractionResponse,
        ):
            logger.info("Extracted function arguments", arguments=arguments)
            return arguments.arguments

        logger.warning("Failed to parse argument extraction response")

        return None

    async def function_calling(
        self,
        *,
        messages: list[Message],
        functions: list[Function],
        few_shot_examples: list[FewShotExample] | None = None,
        mode: Mode = Mode.Json,
        llm_options: dict[str, Any] | None = None,
    ) -> FunctionCallingOutput | None:
        # Select function.
        function_name = await self.select_function(
            messages=messages,
            functions=functions,
            few_shot_examples=few_shot_examples,
            llm_options=llm_options,
        )

        if isinstance(function_name, NotSelected):
            return FunctionCallingOutput(function=NOT_SELECTED)

        if not function_name:
            return None

        function_obj = next(
            (x for x in functions if x.name == function_name),
            None,
        )
        if not function_obj:
            logger.warning(
                "LLM generated invalid function",
                invalid_function=function_name,
            )
            return None

        logger.info("Selected function", function_name=function_name)

        # Extract arguments.
        arguments = await self.extract_arguments(
            messages=messages,
            function=function_obj,
            few_shot_examples=few_shot_examples,
            mode=mode,
            llm_options=llm_options,
        )
        if arguments is None:
            return None

        logger.info("Extracted arguments", arguments=arguments)

        return FunctionCallingOutput(
            function=SelectedFunction(
                name=function_name,
                arguments=json.dumps(arguments),
            ),
        )
